from strenum import StrEnum

class Enums(StrEnum):
    
    STATUS_FAILED = "failed"
    COUNTRY_NOT_WHITELISTED = "this country is not whitelisted"
    INVALID_API_KEY = "invalid api key"
    INSUFFICIENT_FUNDS = "you have no balance"
    MISSING_PARAMETERS = "you are missing parameters"


from uuid import uuid4
from hashlib import sha256
from strenum import StrEnum

class Enums(StrEnum):
    RINGING = "call.ringing"
    ANSWERED = "call.answered"
    HANGUP = "call.ended"
    ENDED = "call.ended"

    DTMF_GATHERED = "dtmf.gathered"
    DTMF_RECIEVED = "Single DTMF"

    VOICEMAIL_DETECTED = "machine.detected"
    HUMAN_DETECTED = "human.detected"
    NOTSURE_DETECTED = "notsure.detected"


class CountryNotWhitelistedException(Exception):
    def __init__(self):
        super().__init__("The country you are trying to dial `from` or `to` is not whitelisted.")

class NonExistentCallException(Exception):
    def __init__(self, debug = False, callSid = ""):
        super().__init__(f"You cannot control a non-existent call or channel.{callSid if debug else ''}")

class WAFTriggeredException(Exception):
    def __init__(self):
        super().__init__("WAF triggered due to suspicious activity from IP address.")

class InvalidAPIKeyException(Exception):
    def __init__(self):
        super().__init__("Invalid API Key")

class InsufficientBalanceException(Exception):
    def __init__(self):
        super().__init__("API key has insufficient balance.")

class MissingParametersException(Exception):
    def __init__(self):
        super().__init__("API Request is malformed, re-check function parameters.")
    
class InvalidDataTypeException(Exception):
    def __init__(self, varName, type):
        super().__init__(f"Var[{varName}] must be data type of {type}.")

class InvalidRegexResultException(Exception):
    def __init__(self):
        super().__init__(f"Value must consist of: digits, '*' or '#'.")


from strenum import StrEnum

class Enums(StrEnum):
    BASE = "https://api.jokerapi.co/voice/v1/"
    
    DIAL = "https://api.jokerapi.co/voice/v1/dial"
    HANGUP = "https://api.jokerapi.co/voice/v1/hangup"
    PLAY_AUDIO = "https://api.jokerapi.co/voice/v1/play"
    PLAY_TEXT = "https://api.jokerapi.co/voice/v1/playtext"
    GATHER_DTMF_WITH_AUDIO = "https://api.jokerapi.co/voice/v1/gather"
    GATHER_DTMF_WITH_TEXT = "https://api.jokerapi.co/voice/v1/gathertext"
    SEND_DTMF = "https://api.jokerapi.co/voice/v1/senddtmf"


import re
from requests import get
from ..enums import *
from ..format import __dial_format__, __hangup_format__, __play_audio_format__, __play_text_format__, __play_audio_dtmf__, __play_text_dtmf__, __send_dtmf_format__
from typing import Any
from ..exceptions import exceptions

class outbound_calls:
    def __init__(self, apiKey: str) -> None:
        self.apiKey = apiKey
    
    def parse_request_response(self, requestResponse: str) -> Any | bool :
        exceptionEnum = {
            response.Enums.COUNTRY_NOT_WHITELISTED: exceptions.CountryNotWhitelistedException,
            response.Enums.INVALID_API_KEY: exceptions.InvalidAPIKeyException,
            response.Enums.INSUFFICIENT_FUNDS: exceptions.InsufficientBalanceException,
            response.Enums.MISSING_PARAMETERS: exceptions.MissingParametersException
        }

        if requestResponse.status_code == 403:
            raise exceptions.WAFTriggeredException()
        
        elif requestResponse.json()['status'] == response.Enums.STATUS_FAILED:
            raise exceptionEnum[requestResponse.json()['context']]()
        
        return False
    
    def create_call(self, to: str, from_: str, callbackUrl: str) -> list[bool, dict[Any]]:
        if self.__retrieve__:
            return
        
        apiResponse = get(__dial_format__(self.apiKey, endpoint.Enums.DIAL, to, from_, callbackUrl))
        
        if not self.parse_request_response(apiResponse):
            return True, apiResponse.json()['callsid']


    def hangup_call(self) -> Any:
        if not self._sid[0]:
            raise exceptions.NonExistentCallException()
        
        self._sid[0] = False
        return get(__hangup_format__(self.apiKey, endpoint.Enums.HANGUP, self.sid))

    def play_(self, _: str, audioUrl_: str = None, text_: str = None, voice_: str = None) -> str:
        if not self._sid[0]:
            raise exceptions.NonExistentCallException()

        formatEnum = {
            "audio": [__play_audio_format__, {"apiKey": self.apiKey, "baseUrl": endpoint.Enums.PLAY_AUDIO.value, "sid": self.sid, "audioUrl": audioUrl_}],
            "text": [__play_text_format__, {"apiKey": self.apiKey, "baseUrl": endpoint.Enums.PLAY_TEXT.value, "sid": self.sid, "text": text_, "voice": voice_}]
        }
        
        return get(formatEnum[_][0](**formatEnum[_][1]))
    
    def gather_dtmf_with(self, _: str, maxDigits_: int, audioUrl_: str = None, text_: str = None, voice_: str = None) -> str:
        if not self._sid[0]:
            raise exceptions.NonExistentCallException()
        
        elif not isinstance(maxDigits_, int):
            raise exceptions.InvalidDataTypeException("maxDigits", "int")
          
        formatEnum = {
            "audio": [__play_audio_dtmf__, {"apiKey": self.apiKey, "baseUrl": endpoint.Enums.GATHER_DTMF_WITH_AUDIO, "sid": self.sid, "audioUrl": audioUrl_, "maxDigits": maxDigits_}],
            "text": [__play_text_dtmf__, {"apiKey": self.apiKey, "baseUrl": endpoint.Enums.GATHER_DTMF_WITH_TEXT, "sid": self.sid, "text": text_, "voice": voice_, "maxDigits": maxDigits_}]
        }

        return get(formatEnum[_][0](**formatEnum[_][1]))

    def send_dtmf_(self, input: Any) -> str:
        if not self._sid[0]:
            raise exceptions.NonExistentCallException()
        
        elif not re.match(r"^[0-9*#]+$", input):
            raise exceptions.InvalidRegexResultException()
        
        return get(__send_dtmf_format__(self.apiKey, endpoint.Enums.SEND_DTMF, self.sid, input))



def __dial_format__(apiKey, baseUrl: str, to: str, from_: str, callbackUrl: str) -> str:
    return f"{baseUrl}?apikey={apiKey}&to={to}&from={from_}&webhookurl={callbackUrl}"

def __hangup_format__(apiKey, baseUrl: str, sid: str) -> str:
    return f"{baseUrl}?apikey={apiKey}&callsid={sid}"

def __play_audio_format__(apiKey, baseUrl: str, sid: str, audioUrl: str) -> str:
    return f"{baseUrl}?apikey{apiKey}&callsid={sid}&audiourl={audioUrl}"

def __play_text_format__(apiKey, baseUrl: str, sid: str, text: str, voice: str) -> str:
    return f"{baseUrl}?apikey={apiKey}&callsid={sid}&text={text}&voice={voice}"

def __play_audio_dtmf__(apiKey, baseUrl: str, sid: str, audioUrl: str, maxDigits: int) -> str:
    return f"{baseUrl}?apikey={apiKey}&callsid={sid}&audiourl={audioUrl}&maxdigits={maxDigits}"

def __play_text_dtmf__(apiKey, baseUrl: str, sid: str, text: str, voice: str, maxDigits: int) -> str:
    return f"{baseUrl}?apikey={apiKey}&callsid={sid}&text={text}&voice={voice}&maxdigits={maxDigits}"

def __send_dtmf_format__(apiKey, baseUrl: str, sid: str, input_) -> str:
    return f"{baseUrl}?apikey={apiKey}&callsid={sid}&input={input_}"


from threading import Thread
from flask import Flask
from waitress import serve
from ..endpoints import outbound_calls

"""
# `create_outbound_call` class to initiate an outbound call.
------------------------------------------------------------
```
JokerObject = create_outbound_call(
    apiKey = "<APIKEY>": Default
    *arg, **kwargs: Required
    >   to: str ~ The phone number to dial.
        from_: str ~ The phone number to dial `to` from.
        callbackUrl: str ~ Any webserver to handle live callbacks to.
)
"""
class create_outbound_call(outbound_calls):
    """
    Main initiator for the `create_outbound_call` class.

    Parameters.
    ----------------------------------------------------
    `apiKey` > The API Key to use the SDK.

    *args > Not required key word arguments which may be used in the future.

    **kwargs > Required arguments to initiate a outbound call.


    `self.sid`: list[int] or list[int, str] 
        > Index 0 representing the status of the call (if it is live or not.): Static

        > Index 1 representing the SID of a channel if Index 0 is True (live).
    """
    def __init__(self, apiKey: str = "<API_KEY>", __retrieve__: bool = False, *args, **kwargs) -> None:
        self.sid: str = ""
        self.__retrieve__ = __retrieve__
        self._sid: list[bool] = [False if not __retrieve__ else True]
        super().__init__(apiKey)
        if not __retrieve__:
            self.__dial__(**kwargs)
    
    """
    'Private' `__dial__` function which is called whenever the class is initiated. 
    """
    def __dial__(self, to: str = "", from_: str = "", callbackUrl: str = "https://0.0.0.0/callbacks") -> str:
        self._sid[0], self.sid = self.create_call(to, from_, callbackUrl)
    
    """
    The `play_audio` function to play live audio into a live channel from a URL.
    """
    def play_audio(self, audioUrl: str = "http://0.0.0.0/audio.wav") -> None:
        return self.play_("audio", audioUrl_ = audioUrl)
    
    """
    The `play_text` function to play text into a live channel without the need of TTS.
    """
    def play_text(self, text: str = "This is an example phrase to play", voice: str = "ai3-en-US-Madison") -> None:
        return self.play_("text", text_ = text, voice_ = voice)
    
    """
    The `gather_dtmf_with_audio` function to play live audio into a live channel from a URL
        Whilst gathering DTMF inputs from the recipient.
    """
    def gather_dtmf_with_audio(self, audioUrl: str = "http://0.0.0.0/audio.wav", maxDigits: int = 0) -> None:
        return self.gather_dtmf_with("audio", audioUrl_ = audioUrl, maxDigits_ = maxDigits)
    
    """
    The `gather_dtmf_with_text` function to play text into a live channel without the need of TTS
        Whilst gathering DTMF inputs from the recipient.
    """
    def gather_dtmf_with_text(self, text: str = "This is an example phrase to play", voice: str = "ai3-en-US-Madison", maxDigits: int = 0) -> None:
        return self.gather_dtmf_with("text", text_ = text, voice_ = voice, maxDigits_ = maxDigits)
    
    """
    The `send_dtmf` function to send DTMF tones to a live channel.
    """
    def send_dtmf(self, input_: str = "123456789*#") -> None:
        return self.send_dtmf_(input_)
    
    """
    The `hangup` function to hangup a live channel.
    """
    def hangup(self) -> None:
        return self.hangup_call()

class retrieve(create_outbound_call):
    def __init__(self, apiKey: str = "<API_KEY>", sid: str = "callSID") -> None:
        super().__init__(apiKey=apiKey, __retrieve__ = True)
        self.sid = sid

class callback_server:
    def __init__(self, app: Flask, host: str = "0.0.0.0", port: int = 9191) -> None:
        self.app: Flask = app
        self.credentials: list[str, int] = [host, port]
    
    def app(self) -> Flask:
        return self.app

    def addCallbackEndpoint(self, function, endpointName: str = "/JokerSDK/callbacks", flagName: str = "The default callback endpoint", requestMethods: list = ["POST", "GET"]) -> None:
        return self.app.add_url_rule(endpointName, flagName, function, methods=requestMethods)
    
    def createCallbackServer(self) -> int | None:
        return Thread(serve(self.app, host = self.credentials[0], port = self.credentials[1])).start()
